/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package beaconconfig;

import gnu.getopt.*;
import java.awt.*;
import javax.swing.border.*;
import javax.swing.*;

import cricketdaemon.clientlib.*;
import cricketdaemon.clientlib.data.*;
import cricketdaemon.clientlib.sim.*;
import java.io.*;
import java.util.*;

public class BeaconConfig implements Callback
{
    public static Random r = new Random(0);

    SolverPanel solver;
    static int numBeacons = 1;
    private ArrayList beaconSelect = null;
    static boolean remoteEnable = false;

    // window dimensions in pixels
    private static final int WIN_SIZE     = 600; // always a square window
    private static final int H_WIN_MARGIN = 10;
    //private static final int V_WIN_MARGIN = 151;
    private static final int V_WIN_MARGIN = 129;
    
    // coordinate system params
    private static final int COORD_WIDTH = 300;
    private static final int GRID_STEP = 61; // 61cm is about 2 feet

    static ServerBroker server;

    SketchPanel sketchPanel;
    Controls controlPanel;

    public BeaconConfig(ServerBroker s,
			int winSize, int coordWidth, int gridStep,
			int numBeacons, int closeness, int stddev, 
			int enoughsamples, 
			int streamerSamplingPeriod,
			int streamerDelay,
			ArrayList beaconSelect)
    {
	server = s;

	//CustomFrame f = new CustomFrame("Beacon Configuration", 410, 513);
	CustomFrame f = new CustomFrame("Beacon Configuration", 
					winSize+H_WIN_MARGIN, winSize+V_WIN_MARGIN);
	Container container = f.getContentPane();
	BorderLayout layout = new BorderLayout();
	container.setLayout(layout);
	f.setBackground(Color.white);

	DrawingArea d = new DrawingArea(winSize, coordWidth, gridStep);
	d.setBorder(createBorder());
	d.setStreamerSamplingPeriod(streamerSamplingPeriod);
	d.setStreamerDelay(streamerDelay);
	container.add(d, "Center");
	
	if(remoteEnable) {
	    NetworkServer ns = new NetworkServer();
	    solver = new RemoteSolverPanel(this, d, closeness, stddev, enoughsamples, ns);
	    sketchPanel = new RemoteSketchPanel(solver, ns, false);
	    controlPanel = new RemoteControls(d, solver, ns, false);
	    ns.addActionListener(sketchPanel);
	    ns.addActionListener(controlPanel);
	    ns.start();
	}
	else {
	    solver = new SolverPanel(this, d, closeness, stddev, enoughsamples);
	    sketchPanel = new SketchPanel(solver);
	    controlPanel = new Controls(d, solver);
	}

	solver.setBorder(createBorder());
	container.add(solver, "South");
	Controls.getResetButton().addActionListener(sketchPanel);

	JPanel cPanel = new JPanel(new BorderLayout(), true);
	cPanel.add(controlPanel, "North");
	cPanel.add(sketchPanel, "South");
	cPanel.setBorder(createBorder());
	container.add(cPanel, "North");

	d.initialize();

	f.validate();
	f.pack();
	f.setVisible(true);

	this.numBeacons = numBeacons;
	this.beaconSelect = beaconSelect;
    }

    public static Border createBorder() {
	Border raisedbevel = BorderFactory.createRaisedBevelBorder();
	Border loweredbevel = BorderFactory.createLoweredBevelBorder();
	
	return
	    BorderFactory.createCompoundBorder(raisedbevel, loweredbevel);
    }


    private ArrayList selectBeacon(ArrayList beaconsHeard) 
    {
	if(beaconSelect == null)
	    return beaconsHeard;

	Iterator it = beaconsHeard.iterator();
	while(it.hasNext()) {
	    BeaconRecord br = (BeaconRecord) it.next();
	    if(! beaconSelect.contains(br) )
		it.remove();
	}
	
	//System.out.println("BeaconConfig::parseBeaconSelect: "+beaconsHeard);
	return beaconsHeard;
    }

    synchronized public void callback(CricketData data, BitSet mask)
    {
	if(mask.get(Broker.BEACONSHEARDLIST)) {
	    solver.processDistStats(selectBeacon(data.getBeaconsHeard()));
	}
	if(mask.get(Broker.DEVICECOORD) || mask.get(Broker.POSERR)) {
	    solver.processDeviceCoord(data.getPosAlgor(), data.getDevicePosition());
	}
    }

    public static void usage()
    {
	System.err.println("Usage: java beaconconfig.BeaconConfig [-n numBeacons][-c closeness][-s stddev_thresh]\n"+
			   "                                      [-g gridStep][-r][-W winSize][-w maxcoord]\n"+
			   "                                      [-e enoughsamples][-d streamerDelay]\n"+
			   "                                      [-t streamerSamplingPeriod]\n");
	System.exit(1);
    }

    private static ArrayList parseBeaconSelect(String s)
    {
	if(s.equals("0"))
	    return null;

	ArrayList result = new ArrayList();
	StringTokenizer st = new StringTokenizer(s, ",");
	while(st.hasMoreTokens()) {
	    BeaconRecord br   = new BeaconRecord();
	    String spaceAndId = st.nextToken();
	    int delim         = spaceAndId.indexOf("%");
	    br.space      = spaceAndId.substring(0,delim);
	    br.id         = Integer.valueOf(spaceAndId.substring(delim+1));
	    br.setUniqueName();
	    result.add(br);
	}
	System.out.println("BeaconConfig::parseBeaconSelect: "+result);
	return result;
    }

    public static void main(String args[]) {
	int enoughsamples = 0; // disabled
	int numBeacons = 3;
	int closeness  = 100; // in cm
	int stddev     = 10;  // in cm
	int winSize    = WIN_SIZE;
	int coordWidth = COORD_WIDTH;
	int gridStep   = GRID_STEP;
	int streamerSamplingPeriod = 200;
	int streamerDelay = 3000;
	ArrayList beaconSelect = null;

	Getopt g = new Getopt("BeaconConfig", args, "B:d:c:e:g:n:rs:t:W:w:");

	int c;
	try {
	    while((c = g.getopt()) != -1) {
		switch(c) {
		case 'B':
		    beaconSelect = parseBeaconSelect(g.getOptarg());
		    break;
		case 'c':
		    closeness = Integer.parseInt(g.getOptarg());
		    break;
		case 'd':
		    streamerDelay = Integer.parseInt(g.getOptarg());
		    break;
		case 'e':
		    enoughsamples = Integer.parseInt(g.getOptarg());
		    break;
		case 'g':
		    gridStep   = Integer.parseInt(g.getOptarg());
		    break;
		case 'n':
		    numBeacons = Integer.parseInt(g.getOptarg());
		    break;
		case 'r':
		    System.err.println("Remote Enabled:  If GUI hangs, check network "+
				       "connectivity\nand restart application\n");
		    remoteEnable = true;
		    break;
		case 's':
		    stddev = Integer.parseInt(g.getOptarg());
		    break;
		case 't':
		    streamerSamplingPeriod = Integer.parseInt(g.getOptarg());
		    break;
		case 'W':
		    winSize = Integer.parseInt(g.getOptarg());
		    break;
		case 'w':
		    coordWidth = Integer.parseInt(g.getOptarg());
		    break;
		default:
		    usage();
		}
	    }
	}
	catch (Exception e) {
	    usage();
	}
	
	if(beaconSelect != null)
	    numBeacons = beaconSelect.size();

	Util.assert_nonjava(numBeacons >= 3, "must specify numBeacons >= 3");

	ServerBroker server = new ServerBroker();	
	BeaconConfig bc = new BeaconConfig(server, winSize, coordWidth, gridStep,
					   numBeacons, closeness, stddev, 
					   enoughsamples,  
					   streamerSamplingPeriod,
					   streamerDelay,
					   beaconSelect);


	BitSet mask = new BitSet();
	mask.set(Broker.BEACONSHEARDLIST);
	mask.set(Broker.DEVICECOORD);
	mask.set(Broker.POSERR);
	server.register(bc, mask);
	server.start();

	try { server.join(); } catch (InterruptedException ie) { }
    }
}
