/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
#include "config.h"
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>

#if defined (HAVE_PATH_H)
#include <paths.h>
#else
#if !defined (_PATH_DEVNULL)
#define _PATH_DEVNULL    "/dev/null"
#endif
#endif

#if defined (HAVE_STRINGS_H)
#include <strings.h>
#endif


#if defined (HAVE_SYS_TERMIOS_H)
#include <sys/termios.h>
#else
#if defined (HAVE_TERMIOS_H)
#include <termios.h>
#endif
#endif

#if defined (HAVE_SYS_SELECT_H)
#include <sys/select.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <stdio.h>

#include "cricketd.h"
#include "version.h"

#define QLEN		5
#define BUFSIZE		2048

int debug = 0;
int device_speed = B115200;
int default_baud = 115200;
int device_type = DEVICE_MIT_CRICKET_V2;
char *device_name = 0;
/* command line option defaults */
char *default_device_name = "/dev/ttyS0";

int nfds;
int reopen = 0;

static int handle_cricket_input(int input, fd_set * afds, fd_set * reg_fds);
static int handle_cricket_v1_1_input(int input, fd_set * afds, fd_set * reg_fds);
static int handle_cricket_v2_input(int input, fd_set * afds, fd_set * reg_fds, fd_set *newfds);
static int handle_request(int serialfd, int fd, fd_set * fds, fd_set *newfds);

static void onsig(int sig)
{
    serial_close();
    syslog(LOG_NOTICE, "Received signal %d. Exiting...", sig);
    exit(10 + sig);
}

static void sigusr1(int sig)
{
  reopen = 1;
}

int daemonize()
{
    int fd;
    pid_t pid;

    pid = fork();

    switch (pid) {
    case -1:
	return -1;
    case 0:
	break;
    default:
	_exit(pid);
    }

    if (setsid() == -1)
	return -1;
    chdir("/");
    fd = open(_PATH_DEVNULL, O_RDWR, 0);
    if (fd != -1) {
	dup2(fd, STDIN_FILENO);
	dup2(fd, STDOUT_FILENO);
	dup2(fd, STDERR_FILENO);
	if (fd > 2)
	    close(fd);
    }
    return 0;
}

#ifdef __CYGWIN__
static void 
cygwin_check(char *device)
{
    fprintf(stderr, "Warn: if cricketd crashes, install cygwin ver \
1.5.10 or later\n");
#define PORTPREFIX "COM"
    if(strncasecmp(device, PORTPREFIX, strlen(PORTPREFIX)) == 0) {
      fprintf(stderr, "\"COM\" ports no longer used since cygwin v.1.5.10\n\
Instead use /dev/ttySx (where x is a decimal digit)\n");
      exit(1);
    }
}
#endif

int main(int argc, char *argv[])
{
    char *default_service = "cricketd";
    char *service = 0;

    struct sockaddr_in fsin;
    int msock;
    fd_set rfds;
    fd_set afds;
    fd_set reg_fds;
    fd_set newfds;
    int alen;
    int baud = default_baud;
    int fd, input;
    int need_cricket;
    extern char *optarg;
    int option;
    int daemondisable = 0;
    char buf[BUFSIZE];

    printf("cricketd version %s\n", VERSION);

    while ((option = getopt(argc, argv, "D:L:S:T:hncl:p:s:d:r:z")) != -1) {
	switch (option) {
	case 'z':
	    daemondisable = 1;
	    break;
	case 'T':
	    switch (*optarg) {
	    case 'c':
  	        device_type = DEVICE_MIT_CRICKET_V1;
		break;
	    case 'k':
  	        device_type = DEVICE_MIT_CRICKET_V1_1;
		break;
	    case 't':
  	        device_type = DEVICE_MIT_CRICKET_V2;
		break;

	    default:
		fprintf(stderr, "Invalid device type \"%s\"\n", optarg);
  	        exit(1);
		break;
	    }
	    break;
	case 'D':
	    debug = (int) strtol(optarg, 0, 0);
	    break;
	case 'S':
	    service = optarg;
	    break;
	case 'p':
	    device_name = optarg;
	    break;
	case 's':
	    baud = atoi(optarg);
	    if (baud < 1200) {	baud = 0; device_speed = B0; }
	    else if (baud < 2400) { baud = 1200; device_speed = B1200; }
	    else if (baud < 4800) { baud = 2400; device_speed = B2400; }
	    else if (baud < 9600) { baud = 4800; device_speed = B4800; }
	    else if (baud < 19200) { baud = 9600; device_speed = B9600; }
	    else if (baud < 38400) { baud = 19200; device_speed = B19200; }
	    else if (baud < 57600) { baud = 38400; device_speed = B38400; }
	    else if (baud < 115200) { baud = 57600; device_speed = B57600; }
	    else { baud = 115200; device_speed = B115200; }
	    break;

	case 'h':
	case '?':
	default:
	    fprintf(stderr, "usage:  cricketd [options] \n\
  options include: \n\
  -D integer   [ set debug level ] \n\
  -S integer   [ set port for daemon (default=%s)] \n\
  -T c         [ MIT Cricket V1 support ] \n\
  -T k         [ MIT Cricket V1.1 support ] \n\
  -T t         [ MIT Cricket V2 support (default)] \n\
  -h           [ help message ] \n\
  -p string    [ set Cricket device name (default=%s)] \n\
  -s baud_rate [ set baud rate on Cricket device (default=%d)] \n\
  -z           [ do not put cricketd into background ] \n\
", DEFAULTPORT, default_device_name, default_baud);
	    exit(0);
	}
    }

    if (!device_name)
	device_name = default_device_name;
    if (!service)
      if (!getservbyname(default_service, "tcp")) {
	service = DEFAULTPORT;
      } else {
	service = default_service;
      }

    fprintf(stderr, "MIT CRICKET SUPPORT (v%s)\n", typetoa(device_type));

    fprintf(stderr, "command line options:\n");

    if (debug > 0)
	fprintf(stderr, "  debug level:    %d\n", debug);

    fprintf(stderr, "  device name:    %s\n", device_name);
    fprintf(stderr, "  device speed:   %d (%d)\n", baud, device_speed);
    fprintf(stderr, "  port:           %s\n", service);

    if (debug < 2 && !daemondisable)
	daemonize();

    /* Handle some signals */
    signal(SIGUSR1, sigusr1);
    signal(SIGINT, onsig);
    signal(SIGHUP, onsig);
    signal(SIGTERM, onsig);
    signal(SIGQUIT, onsig);
    signal(SIGPIPE, SIG_IGN);

    openlog("cricketd", LOG_PID | LOG_PERROR, LOG_USER);
    syslog(LOG_NOTICE, "cricketd started (version %s)", VERSION);
    syslog(LOG_NOTICE, "cricketd listening on port %s", service);

    msock = passiveTCP(service, QLEN);

#ifdef __CYGWIN__
    cygwin_check(device_name);
    nfds = 32;
#else
    nfds = getdtablesize();
#endif

    FD_ZERO(&afds);
    FD_ZERO(&reg_fds);
    FD_ZERO(&newfds);
    FD_SET(msock, &afds);

    input = -1;

    while (1) {
	// afds contains registered fds
	// rfds contains triggered fds
        memcpy((char *)&rfds, (char *)&afds, sizeof(rfds));

	if (select(nfds, &rfds, (fd_set *) 0, (fd_set *) 0,
		   (struct timeval *) 0) < 0) {
	  if (errno == EINTR)
		continue;
	    errexit("select");
	}

	if (FD_ISSET(msock, &rfds)) {
	    int ssock;

	    alen = sizeof(fsin);
	    ssock = accept(msock, (struct sockaddr *) &fsin,
			   &alen);

	    if (ssock < 0)
		errlog("accept incoming TCP connection failed");
	    else {
		FD_SET(ssock, &afds);
	    }
	}

	if (input >= 0 && FD_ISSET(input, &rfds)) {
	    switch (device_type) {
	    case DEVICE_MIT_CRICKET_V1:
  	        handle_cricket_input(input, &afds, &reg_fds);
		break;
	    case DEVICE_MIT_CRICKET_V1_1:
		/* Crickets send ascii strings */
		handle_cricket_v1_1_input(input, &afds, &reg_fds);
		break;
	    case DEVICE_MIT_CRICKET_V2:
		handle_cricket_v2_input(input, &afds, &reg_fds, &newfds);
		break;
	    }
	}

	need_cricket = 0;
	if (reopen) {
	  FD_CLR(input, &afds);
	  input = -1;
	  serial_close();
	  syslog(LOG_NOTICE, "Closed Cricket Serial Port");
	  need_cricket++;
	}

	/* process TCP client input */
	for (fd = 0; fd < nfds; fd++) {
	    if (fd != msock && fd != input) {

		if(FD_ISSET(fd, &rfds)) {
		    if (input == -1) {
			if ((input = serial_open()) < 0)
			    errexit("serial open: ");
			syslog(LOG_NOTICE, "Opened Cricket Serial Port");
			FD_SET(input, &afds);
		    }
		    if (handle_request(input, fd, &reg_fds, &newfds) == 0) {
			(void) close(fd);
			FD_CLR(fd, &afds);
			FD_CLR(fd, &reg_fds);
		    }
		}
		
		// if there is a registered client...
		if(FD_ISSET(fd, &afds))
		    need_cricket++;
	    }
	}

	/* when there are no TCP clients, we can close the serial port */
	if (!need_cricket && input != -1) {
	    FD_CLR(input, &afds);
	    input = -1;
	    serial_close();
	    syslog(LOG_NOTICE, "Closed Cricket Serial Port");
	}
    }
}

/* 
 * Handle input from the telnet side, not the Cricket side.
 */
static int handle_request(int serialfd, int fd, fd_set * fds, fd_set *newfds)
{
    char buf[BUFSIZE];
    char reply[BUFSIZE];
    char *p;
    int cc;
    int sw;

    cc = read(fd, buf, sizeof(buf) - 1);
    if (cc < 0)
	return 0;

    bzero(reply, sizeof(reply));
    buf[cc] = '\0';
    p = buf;

    if(p[0] == 'R' || p[0] == 'r') {
	if (FD_ISSET(fd, fds)) {
	    FD_CLR(fd, fds);
	    FD_CLR(fd, newfds);
	    sprintf(reply,"\r\nDeregister (Disable)");
	} else {
	    FD_SET(fd, fds);
	    FD_SET(fd, newfds);
	    sprintf(reply, "\r\nRegister (Enable)");
	}
	strcat(reply, "\r\n");

	if (write(fd, reply, strlen(reply)) < 0) {
	  return 0;
	}
    }
    else if (serialfd >= 0) {

	if(cc >= 2) {
	    // CricketM.nc expects only one \n and or one \r,
	    // not both \r\n.  So take out the \r if \r\n is 
	    // present.
	    if(buf[cc-1] == '\n' && buf[cc-2] == '\r') {
		buf[cc-2] = '\n';
		buf[cc-1]   = '\0';
		cc -= 1;
	    }
	}
	sw = write(serialfd, buf, cc);
	//sprintf(reply+strlen(reply), "[%d]:%s", sw, buf);
	//sprintf(reply, "%s", buf);
    }

    return cc;
}

#define FIELD_SIZE 256
#define OUTBUF_SIZE 8096

static void parse_string(unsigned char* buf, int length, 
			 unsigned char* output, int offset)
{
  if(offset + length + 1 >= OUTBUF_SIZE)
    return; // too large for output

  bcopy(buf, output+offset, length);
  output[offset+length] = '\0';
}

static void parse_int(unsigned char* buf, int length, 
		      unsigned char* output, int offset)
{
    unsigned char my_buf[64];
    unsigned int result;

    if(length < 1 || length > 4)
	return;  // invalid length

    result = buf[0];
	
    if(length > 1)
	result = buf[1] + result*256;
    if(length > 2)
	result = buf[2] + result*256;
    if(length > 3)
	result = buf[3] + result*256;

    sprintf(my_buf, "%x", result);
    if(offset + strlen(my_buf) + 1 >= OUTBUF_SIZE)
	return; // too large for output
    
    strcpy(output+offset, my_buf);
}

static void parse_coord(unsigned char* buf, int length, 
			unsigned char* output, int offset)
{
    return; // not implemented yet
}

static void parse_phase(unsigned char* buf, int length, 
			unsigned char* output, int offset)
{
    return; // not implemented yet
}

static char* field_type_map[] = 
{ 
    "null", 
    "space", 
    "id", 
    "dist", 
    "coord", 
    "phase" 
};

typedef void (*field_parser_fn)(unsigned char*, int, unsigned char*, int);

// need function pointers
static field_parser_fn field_parser[] =
{ 
    0, 
    &parse_string, 
    &parse_int, 
    &parse_int, 
    &parse_coord, 
    &parse_phase 
};

static void parseField(unsigned char* buf, int field_length, 
		       unsigned char* output) 
{
    int field_id;
    int cur_length, new_length; 

    // subtract length of field_id to get acutal length of the value field
    field_length -= 1; 
    if(field_length < 0)
      return; // invalid field 
    field_id = buf[0]; //((buf[0] & 0xff) << 8) | (buf[1] & 0xff);
    
    if(field_id < 0 || field_id > 5) return;
    
    cur_length = strlen(output);
    
    // cur_length + label length + comma + equal sign
    new_length = cur_length + strlen(field_type_map[field_id]) + 2;
    
	// assumes that field_type do not contain any special characters
    if(new_length + 1 >= OUTBUF_SIZE) return; // include null char
    
    //printf("parseField cur_length=%d field_length=%d field_id=%d\n", 
    //cur_length, field_length, field_id);
    sprintf(output + cur_length, ",%s=", field_type_map[field_id]);
    field_parser[field_id](&buf[1], field_length, output, new_length);
    //    printf("parseField-after: %s\n", output);
}

static 
int parse_cricket_packet(fd_set * afds, fd_set * reg_fds,
			 unsigned char buf[], int buf_off, int end_index)
{
  unsigned char field_buf[FIELD_SIZE];
  unsigned char output[OUTBUF_SIZE];
  int offset = 0;
  int unstuffed = 0; //tells if a leading 0 was removed
  int fd;
  
  memset(output, 0, OUTBUF_SIZE*sizeof(unsigned char));
  sprintf(output, "$Cricket2"); 
  
  if(end_index - buf_off < 2)
    return 0;
  
  field_buf[0] = buf[buf_off++]; // filling parse window
  field_buf[1] = buf[buf_off++]; // filling parse window
  offset = 2;
  unstuffed = 0;
  
  while (1) {    
    if(offset >= FIELD_SIZE) {
      return 0; // buffer overrun
    }
    
    field_buf[offset] = buf[buf_off];
    offset++; 
    buf_off++;
    
    //check for end of packet
    if(!unstuffed) {
      if(buf_off == end_index + 1) {
	// reached end of packet
	
	// check validity of data (to check the validity, allow for
	// an external module to register with this specifically for
	// error detection/correction by default if there is no
	// registered component, do not do any test
	
	parseField(field_buf, offset-3, output);
	
	if(strlen(output) >= OUTBUF_SIZE)
	  return 1;  // no space to add newline character
	
	strcat(output, "\n");
	
	if(debug >= 7)
	  printf("write: %s\n", output);
	
	// send output to registered apps
	for (fd = 0; fd < nfds; fd++) {
	  if (FD_ISSET(fd, reg_fds)) {
	    if (write(fd, output, strlen(output)) < 0) {
	      printf("writing out to %d\n", fd);
	      syslog(LOG_NOTICE, "Raw write: %s", strerror(errno));
	      FD_CLR(fd, afds);
	      FD_CLR(fd, reg_fds);
	    }
	  }
	}
	
	return 1;
      }//check for new subfield
      else if(( field_buf[offset - 3] == 2 ) && 
	      ( field_buf[offset - 2] == 0 ) &&
	      ( field_buf[offset - 1] == 2 )) { 
	parseField(field_buf, offset-3, output);
	
	//reset
	field_buf[0] = buf[buf_off++]; // filling parse window
	field_buf[1] = buf[buf_off++]; // filling parse window
	offset = 2;
	unstuffed = 0;
      }
      else if ( field_buf[offset-3] == 0 ) { 
	//do byte unstuffing for the byte at offset - 3                
	//check for start of 0 string 
	//remove stuffed 0
	field_buf[offset-3] = field_buf[offset-2];
	field_buf[offset-2] = field_buf[offset-1];                    
	offset--; //remove the stuffed 0
	unstuffed = 1;
      } 
    }
    else if( field_buf[offset - 3] != 0 ) { 
      // check for end of 0 string 
      unstuffed = 0;
    }
  }          
}

// parse the 010101XXXXXXXXX101
static int handle_cricket_input(int input, fd_set * afds, fd_set * reg_fds)
{

  static unsigned char buf[BUFSIZE];
  static int offset = 0;
  static int begin_offset = -1;
  int i = 0;
  
  while(offset < BUFSIZE) {
    // synchronization
    if (read(input, buf+offset, 1) != 1) 
      return 1;
    //printf("handler 1 %d %d\n", offset, begin_offset);
    //printf("read: %x\n", buf[offset]);
    
    if( offset >= 5 && 
	buf[offset] == 1 &&
	buf[offset-1] == 0 &&
	buf[offset-2] == 1 &&
	buf[offset-3] == 0 &&
	buf[offset-4] == 1 &&
	buf[offset-5] == 0 )
      {
	offset++;
	begin_offset = offset;
      }
    else if (offset >= 2 &&
	     buf[offset] == 1 &&
	     buf[offset-1] == 0 &&
	     buf[offset-2] == 1 &&
	     begin_offset >= 0) 
      {
	if(debug >= 7) {
	  printf("handler 3: ");
	  printf("[HEAD] ");
	  for(i = 0; i < begin_offset; i++)
	    printf("%x ", buf[i]);
	  printf("[BODY] ");
	  for(i = begin_offset; i < offset-2; i++)
	    printf("%x ", buf[i]);
	  printf("[TAIL] ");
	  for(i = offset-2; i <= offset; i++)
	    printf("%x ", buf[i]);
	  printf("\n");
	}
	parse_cricket_packet(afds, reg_fds, buf, begin_offset, offset);
	begin_offset = -1;
	offset = 0;
	return 1;
      }
    else 
      {
	offset++;
      }
  }
  
  // buffer full so drop everything and start over.
  begin_offset = -1;
  offset = 0;
  
  return 1;
}


//
// V1_1 Crickets send straight ascii strings, so no parsing needs to
// be done, simply pass along the received string
//
static int handle_cricket_v1_1_input(int input, fd_set * afds, fd_set * reg_fds)
{

  unsigned char buf[BUFSIZE];
  int chars_read;
  int fd;

  
  chars_read = read(input, buf, BUFSIZE-1);
  if (chars_read == 0)
    return 1;

  buf[chars_read] = '\0';  // Terminate string

  if(debug >= 7)
    printf("write: %s\n", buf);
	
  // send output to registered apps
  for (fd = 0; fd < nfds; fd++) {
    if (FD_ISSET(fd, reg_fds)) {
      if (write(fd, buf, chars_read) < 0) {
	printf("writing out to %d\n", fd);
	syslog(LOG_NOTICE, "Raw write: %s", strerror(errno));
	FD_CLR(fd, afds);
	FD_CLR(fd, reg_fds);
      }
    }
  }
	
  return 1;
}

// send output to registered apps
static void send_buf(char *buf, int len, int newline, fd_set * afds, fd_set * reg_fds, fd_set *newfds)
{
    int fd;
    unsigned char temp[BUFSIZE];

    if(debug >= 7) {
	memcpy(temp, buf, len);
	temp[len] = '\0';
	printf("write: %s\n", temp);
	//if(temp[len-1] == '\n' && temp[len-2] == '\r')
	//printf("yes\n");
    }

    for (fd = 0; fd < nfds; fd++) {
	if (FD_ISSET(fd, reg_fds)) {
 	    if(FD_ISSET(fd, newfds) && !newline) {
		//printf("skip %d\n", newline);
	        continue;
	    }

	    //printf("writing out to %d\n", fd);

	    if (write(fd, buf, len) < 0) {
	        printf("writing out to %d\n", fd);
	        syslog(LOG_NOTICE, "Raw write: %s", strerror(errno));
		FD_CLR(fd, afds);
		FD_CLR(fd, reg_fds);
	    }
	    
	    if(newline)
	      FD_CLR(fd, newfds);
	}
    }
}

static void get_timestamp(unsigned char *buf)
{
    struct timeval tv;
    long long time;

    gettimeofday(&tv, 0);

    time = (long long) tv.tv_usec/1000;

    if(tv.tv_usec % 1000 > 500) 
	time++;

    time += ((long long) tv.tv_sec)*1000;

    if(debug >= 7)
	printf("time: %lld %d %d\n", time, tv.tv_sec, tv.tv_usec);

    sprintf(buf, ",LT=%lld", time);
}

static int cricket_v2_data(unsigned char *buf)
{
    const char *head = "VR=2.0";
    return strncmp(buf, head, strlen(head)) == 0;
}
     

//
// V2 Crickets send straight ascii strings like v1_1, except 
// we append an additional local clock timestamp
//
static int handle_cricket_v2_input(int input, fd_set * afds, fd_set * reg_fds, fd_set *newfds)
{

    unsigned char buf[BUFSIZE];
    int chars_read = 0;

    static int startofline = 0;
    static unsigned char tbuf[BUFSIZE];

    int n;
    int last_sent = 0;
    int newline = 0;
    
    chars_read = read(input, buf, BUFSIZE-1);
    if (chars_read == 0)
	return 1;
    
    buf[chars_read] = '\0';  // Terminate string
    
    if(debug >= 7)
	printf("buf[%d]: %s\n", chars_read, buf);

    for(n = 0; n < chars_read; n++) {
	if(!startofline && (newline = cricket_v2_data(buf+n))) {
	    startofline = 1;
	    get_timestamp(tbuf);
	}
	
	if(startofline) {
	    if((buf[n] == '\r') || (buf[n] == '\n')) {
		// endofline

		// comment out to make beacons work.
		// the problem is that beacon output does not begin with VR=
		// prefix.

		//send_buf(buf+last_sent, n, last_sent, afds, reg_fds, newfds);
		//send_buf(tbuf, strlen(tbuf), last_sent, afds, reg_fds, newfds);

		// if last_sent > 0, then last_sent represents the beginning
		// of a line
		//last_sent = n;
		startofline = 0;
	    }
	}
    }


    // comment out to make beacons work.
    // the problem is that beacon output does not begin with VR=
    // prefix.

    // sending remaining bytes that do not end in \r or \n
    //if(chars_read - last_sent > 0) {
    //send_buf(buf+last_sent, chars_read - last_sent, last_sent, afds, reg_fds, newfds);
	//}
    send_buf(buf+last_sent, chars_read - last_sent, 1, afds, reg_fds, newfds);

    return 1;
}

void errlog(char *s)
{
    syslog(LOG_ERR, "%s: %s\n", s, strerror(errno));
}


int errexit(char *s)
{
    syslog(LOG_ERR, "%s: %s\n", s, strerror(errno));
    fprintf(stderr, "%s: %s\n", s, strerror(errno));
    serial_close();
    exit(2);
}










