/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.clientlib;

import cricketdaemon.clientlib.data.*;
import java.util.*;

/**
 * A thread that interfaces between the Application and the 
 * <code>cricketdaemon</code>. Applications that wish to obtain information 
 * from the <code>cricketdaemon</code> MUST implement <code>Callback</code>,
 * and register with the <code>Broker</code> thread.  The <code>Broker</code>
 * thread manges (de)registration of Application threads, and dispatches the 
 * application callback whenever there is information for the Application.
 *
 * An application needs to create only one instance of Broker; multiple
 * threads of the same application should register with the same Broker.
 *
 * <code>Broker</code> is an abstract class.  An application will either use
 * the <code>ServerBroker</code> or the <code>SimBroker</code> to talk to
 * the actual or simulated Broker.
 *
 * @author Allen Miu
 */
public abstract class Broker extends Thread
{
    /**
     * Holds the coordinates (ViewFinder.Coord) of the services
     * and beacons.  These information should be provided by
     * the service discovery mechanism such as INS.  We may remove
     * these variables from the future Broker API because service 
     * discovery should be made orthorgonal to location estimation.
     */

    protected ArrayList serviceCoords;
    protected ArrayList beaconCoords;

    /* info from service discovery system such as INS; 
     * the exact data structure used for this has yet to be fixed.
     */
    public ArrayList requestServiceCoords()
    {
	return serviceCoords;
    }

    /**
     * Contains a list of applications that registered with this modules.
     * An ArrayList of <code>RegistryEntry</code>.
     */
    private ArrayList registry;

    /**
     * Contains an OR of the requested field masks by the registered modules
     */
    protected BitSet registeredMask;

    /**
     * Timeout period to call process(), when inputBuf is empty.  "0"
     * causes thread to wait indefinitely until notify() or notifyAll()
     * is called.
     */
    protected long timeout;

    /*--- Each module should define its own set of available data types. ---*/

    /**
     * Any object that implements <code>Callback</code> and registers with
     * the "ALL" bit set in the fieldMask may get a callback from the
     * Broker whenever it receives ANY message on the UDP port.
     * This mode is mainly useful for debugging.
     */
    public static final int ALL              = 0; 
    public static final int SPACE            = 1; 
    public static final int SEQNUM           = 2;  // will deprecate
    public static final int DEVICECOORD      = 3;
    public static final int ORIENTATION      = 4;
    public static final int ANCHORBEACON     = 5;
    public static final int BEACONSHEARDLIST = 6;
    public static final int BEACONSTAT       = 7;
    public static final int CALCBEACONDIST   = 8;
    public static final int POSERR           = 20;

    protected static BitSet CLEARMASK;

    static {
	CLEARMASK = new BitSet();
	CLEARMASK.set(0);
    }

    /**
     * A RegistryEntry is made to record the state of each registered app.
     */
    private class RegistryEntry
    {
	Callback callee; // object to be called
	BitSet fieldMask;         // specifies desired fields

	private RegistryEntry(Callback c, BitSet t)
	{
	    callee = c; fieldMask = t;
	}

	public boolean isEqual(RegistryEntry e)
	{
	    // Should not use "callee.isEqual(e.callee)"
	    // because we want to ensure the references points
	    // to the same module
	    return (callee == e.callee) && (fieldMask.equals(e.fieldMask));
	}

	public void matchAndFireCallback(CricketData value, BitSet t)
	{
	    BitSet test = (BitSet) t.clone();
	    test.and(fieldMask);
	    if(fieldMask.length() == 0 || test.length() > 0)  // match any
		callee.callback(value, test); 
	}

	public String toString()
	{
	    return callee+" "+fieldMask;
	}
    }

    /**
     * Creates a Broker object with <code>timeout</code>.  If no new data
     * reaches the Broker for a <code>timeout</code> period, a callback is
     * fired (dispatched).  If <code>timeout</code> is zero, then no 
     * callback is fired until new data reaches the Broker.
     */
    public Broker(int timeo) 
    {
	timeout        = timeo;
	serviceCoords = new ArrayList();
	beaconCoords  = new ArrayList();

	registry = new ArrayList();
	// the "ALL" data type is always available
	registeredMask = (BitSet) CLEARMASK.clone();
    }

    /**
     * Registers an application. Application passes a reference to itself
     * and specify the type of information it wish to acquire.  Hence, the
     * callback will be fired only when the type of the received data 
     * matches the specified in fieldMask.  A fieldMask with "ALL" specified
     * will cause the Broker to fire a callback whenever new data is received,
     * regardless of the data type.
     */
    public synchronized void register(Callback app, BitSet fieldMask)
    {
	//System.out.println("register");
	RegistryEntry newEntry = new RegistryEntry(app, fieldMask);
	if(registry.contains(newEntry)) {
	    System.err.println("Broker::register "+
			       app+" multiple entries "+ newEntry);
	    return; // silently ignore
	}
	registeredMask.or(fieldMask);
	registry.add(newEntry);
    }

    /**
     * Deregisters an Application thread.
     */
    public synchronized void deRegister(Callback c, BitSet fieldMask)
    {
	System.out.println("deregister");
	registeredMask = (BitSet) CLEARMASK.clone();
	for(int i = 0; i < registry.size(); i++) {
	    RegistryEntry entry = (RegistryEntry) registry.get(i);
	    if(entry.callee == c && 
	       (fieldMask == null || entry.fieldMask.equals(fieldMask)))
		registry.remove(i);
	    else // save the fieldMask from the remaining entries
		registeredMask.or(entry.fieldMask); 
	}
    }

    /**
     * Used by subclasses to fire callbacks to the registered applications.
     */
    protected void transfer(CricketData value)
    {
	value.getMask().set(ALL); // to enable ALL data type

	for(int i = 0; i < registry.size(); i++) {
	    RegistryEntry e = (RegistryEntry) registry.get(i);
	    e.matchAndFireCallback(value, value.getMask());
	}
    }

    /**
     * Thread to handle incoming data and callbacks.
     */
    abstract public void run();
}
