/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.module;

import cricketdaemon.Util;
/**
 * A subclass of <code>ModuleThread</code> that differs in the way it
 * interprets the <code>timeoutPeriod</code> paramerter in the
 * constructor.
 *
 * In <code>ModuleThread</code>, the time out is the argument to the
 * <code>wait()</code> call within the default <code>run()</code>
 * method.  Hence, the default <code>run()</code> method will call
 * <code>process()</code> if <bold>no</bold> event occurs for
 * <code>timeout</code> period since the last iteration of the
 * <code>while</code> loop inside the default <code>run()</code> method.
 *
 * The <code>PeriodicModuleThread</code> class calls <code>process()</code>
 * in regular periods, despite the number of <code>process()</code> calls 
 * made in the <code>while</code> within the <code>timeoutPeriod</code>.
 *
 * <code>periodicallyProcess()</code> is called once every 
 * <code>timeoutPeriod</code>, while <code>process()</code> can be called
 * asynchronously.
 *
 * @author Allen Miu
 */
abstract public class PeriodicModuleThread extends ModuleThread
{
    private boolean timeExpired;
    private long expiryTime;

    public PeriodicModuleThread()
    {
	expiryTime = System.currentTimeMillis();
    }

    public PeriodicModuleThread(String name, long timeoutPeriod)
    {
	super(name, timeoutPeriod);
	expiryTime = System.currentTimeMillis();
    }

    public boolean isTimerExpired() { return timeExpired; }
    public long getExpiryTime() { 
	if(isTimerExpired())
	    return System.currentTimeMillis() + timeout; // estimate
	else
	    return expiryTime;
    }

    public void run()
    {
	try {
	    while(true) {
		long waitTime = expiryTime - System.currentTimeMillis();
		timeExpired = waitTime <= 0;
		if(!timeExpired) {
		    synchronized(this) {
			wait(waitTime);
		    }
		}
		process();
		if(timeExpired)
		    expiryTime = timeout + System.currentTimeMillis();
	    }    
	} catch (InterruptedException e) {
	    Util.log(getName() +" interrupted");
	}
	moduleCleanUp();
	Util.log(getName() +" exits");
    }
    
    protected void process() 
    {
	super.process();
	if(isTimerExpired())
	    periodicallyProcess();
    }

    abstract protected void periodicallyProcess();
}
