/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.space;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.reader.parse.ParseMap;
import cricketdaemon.reader.*;
import java.util.*;

/**
 * A module that implements the original inference algorithm (presented 
 * in Mobicom 2000) to identify the set of nearest beacons.
 *
 * A sliding window version of this algorithm is being implemented
 * as BeaconStatModule.  We will phase out the Space* modules soon.
 *
 * @author Allen Miu, Bodhi Priyantha
 */
public class SpaceModule extends ModuleThread 
    implements Beacon.BeaconHandler
{
    private static BitSet spaceFieldMask;

    /* Space field types */
    
    public final static int MEAN = 1;
    public final static int MODE = 2;

    /* space inference parameters */

    // The minimum frequency of occurence for a valied mode
    final static int MODETHRESH = 2;

    // No. of minimum data points needed for calculating
    // mean (to prevent a single value giving a wrong result)
    public final static int MEANTHRESH = 4;

    // No. of consistant values for a beacon to have full set of data
    public final static int ENOUGHPOINTS = 4; // should be 4

    // distance measurement precision in inches
    public final static int PRECISION = 10; // should be 10

    // Percentage of full beacons for a full data set
    public final static int DATATHRESH = 75;

    private Hashtable beaconsHeard;
    private int fullBeacons;
    private int beaconCount;

    public SpaceModule(ModuleThread reader)
    {
	super("SpaceModule", 0);

	/* registering with the "port" reader */
	BitSet desiredTypeMask = new BitSet();
	desiredTypeMask.set(ParseMap.SPACE);
	desiredTypeMask.set(ParseMap.BEACONID);
	desiredTypeMask.set(ParseMap.DISTANCE);
	reader.register(this, desiredTypeMask);

	/* reset */
	beaconsHeard = new Hashtable();
	fullBeacons = 0;
	beaconCount = 0;
    }

    private void reset()
    {
	beaconsHeard.clear();
	fullBeacons = 0;
	beaconCount = 0;
    }

    protected void moduleCleanUp() { return; }
    // protected void process(); // using the inherited process method

    public void processBeacon(Beacon newData, BitSet fieldMask)
    {
	DistanceEstimate[] mean = null, mode = null;

	if(!fieldMask.get(ParseMap.SPACE)) {
	    Util.log("SpaceModule::process() space is null");
	    return; // silently drops
	}
	if(!fieldMask.get(ParseMap.DISTANCE)) {
	    Util.log("SpaceModule::process() distance is null");
	    return; // silently drops
	}
	String space      = (String)  newData.getField(ParseMap.SPACE);
	Integer distance  = (Integer) newData.getField(ParseMap.DISTANCE);

	Integer beaconID = null;
	if(fieldMask.get(ParseMap.BEACONID))
	    beaconID  = (Integer) newData.getField(ParseMap.BEACONID);
	else
	    beaconID = new Integer(0);
	
	SpaceData newProcessedData = null;
	spaceFieldMask = (BitSet) CLEARMASK.clone();
	
	if(insert(space, beaconID.intValue(), distance.intValue())) {
	    // we have enough data points to generate some useful info
	    newProcessedData = new SpaceData(newData);
	    if(registeredMask.get(MEAN)) {
		mean = analyze(MEAN);
		if(mean != null) {
		    newProcessedData.insert(mean, MEAN);
		    spaceFieldMask.set(MEAN);
		}
	    }
	    // todo: awefullly ugly to duplicate code like this
	    if(registeredMask.get(MODE)) {
		mode = analyze(MODE);
		if(mode != null) {
		    newProcessedData.insert(mode, MODE);
		    spaceFieldMask.set(MODE);		
		    }
	    }
	}
	
	/* send space data */
	if(newProcessedData != null)
	    transfer(newProcessedData, spaceFieldMask);
	
	if(mean != null && mode != null)
	    reset();
    }

   /**
     * Inserts a new distance value read from the PortReader.
     * Returns true if there are enough full data points.
     */
    public boolean insert(String space, int id, int dist)
    {
	String uniqueID = space+id;
	if(beaconsHeard.containsKey(uniqueID)) { 
	    // if in table, just insert the reading
	    BeaconDistanceEstimator beaconBin = 
		(BeaconDistanceEstimator) beaconsHeard.get(uniqueID); 
	    if( beaconBin.insert(dist) )
		fullBeacons++;
	}
	else { // if not in table, insert a new BeaconDistanceEstimator
	    BeaconDistanceEstimator newData = 
		new BeaconDistanceEstimator(space,id,dist);
	    beaconsHeard.put(uniqueID, newData);  // hash it
	    beaconCount++;
	}

	Util.assert_nonjava(beaconCount > 0, "SpaceModule::insert() beaconCount == 0");
	return hasEnoughData();
    }

    /**
     * Used to determine if enough data is available for analysis
     * Currently this is the same test as used for MINMODE analysis 
     */
    protected boolean hasEnoughData()
    {
	// percentage of full beacons
	//System.out.println(System.currentTimeMillis()+" "+fullBeacons
	//+" "+beaconCount);
	int temp = (int) ( ((float)fullBeacons/(float)beaconCount) *100 ); 
	return (temp >= DATATHRESH);
    }

    /**
     * This analyses the readings using the appropriate algorithm and
     * sorts the beacon-distance output.
     */
    private DistanceEstimate[] analyze(int meanMode)
    {
	ArrayList beaconDistanceEstimates = new ArrayList();

	// run analysis for all the beacons
	Enumeration beacons = beaconsHeard.elements();
	while(beacons.hasMoreElements()) {
	    BeaconDistanceEstimator beacon = 
		(BeaconDistanceEstimator) beacons.nextElement();
	    DistanceEstimate dist = beacon.distance(meanMode);
		
	    // filters out those beacon estimators without enough data pts
	    if(dist != null)
		beaconDistanceEstimates.add(dist);  
	}
	if(beaconDistanceEstimates.size() == 0) {
	    Util.warn("SpaceModule::analysis() not enough data points");
	    return null;
	}
	// sort disData
	return sort(beaconDistanceEstimates);
    } 

    public static DistanceEstimate[] sort(ArrayList estimates) 
    {
	Object[] temp = (Object[]) estimates.toArray();
	Arrays.sort(temp, (DistanceEstimate) temp[0]);

	// Welcome javaland - can't get around type casting....!
	DistanceEstimate[] result = new DistanceEstimate[temp.length];
	System.arraycopy(temp, 0, result, 0, result.length);

	return result;
    }
}

